/* bsmtp.c
**
** send each message in a BSMTP-folder via SMTP to a specified host.
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "prefs.h"
#include "hlsock.h"
#include "vxerror.h"
#include "tools.h"
#include "folder.h"
#include "exithl.h"
#include "procttl.h"
#include "exitcodes.h"
#include "inline-funcs.h"
#include "mrmacs.h"
#include "sigwrapper.h"

#include "vxhost.h"

/* Lies smtliche Zeilen einer SMTP-Statusmeldung ein; gib aber nur den
** Status-Code ser ersten Zeile zurck.
*/
static int smtp_result (HLSock hlc, char **res)
{
    char line[1024], *p;
    int ecode = 0, stv;
    /* 26.5.2000 bj: set the response timeout to the configured value */
    alarm (prefs.bsmtp_timeout);
    while (HLGets (line, 1024, hlc) != NULL) {
        cuteol (line);
        vxlog (4, "%s\n", line);
        stv = (int) strtol (line, &p, 10);
        if (!ecode) { ecode = stv; }
        if (*p == '\r' || *p == '\n' || *p == ' ') {
	    if (res) { *res = gdupl (__FILE__, __LINE__, line); }
	    break;
	}
    }
    alarm (0);
    return ecode;
}

/* Die folgenden Variablen werden fr den (neu hinzugekommenen) `timeout'-
** Mechanismus bentigt. Die Werte werden in `bsmtp' bzw. `SaveFailed' zum
** frhestmglichen Zeitpunkt zugewiesen, damit der Alarmmechanismus richtig
** funktioniert.
** Es sind:
** sbFx		- File-Handle des (zum lesen) geffneten Folders,
** ssav		- File-Handle der (geretteten) unzustellbaren E-Mails,
** current_msg	- Nummer der ersten E-Mail innerhalb des Folders, die nicht
**                mehr (oder nicht mehr vollstndig) verschickt werden konnte.
*/
static xFile sbFx = (xFile) 0, ssav = (xFile) 0;
static int current_msg = 0;

/* SingleStore (from, to)
**     bertrage eine E-Mail aus einem VxMail-Folder in einen anderen.
*/
static __inline__ void SingleStore (xFile src, xFile dst)
{
    char line[1024];
    flCopyUID (src, dst);
    while (xGets (line, 1024, src) > 0) { xPuts (line, dst); }
    xSet (dst);
}

/* MultiStore (from, to, first, savestate)
**     bertrage alle E-Mails ab der Nummer `first' aus einem VxMail-Folder
**     in einen anderen.
*/
static int MultiStore (xFile src, xFile dst, int from, int savestate)
{
    int ix;
    char buf[60];
    for (ix = from; xSeek (src, (off_t) ix) >= 0; ++ix) {
	if (xGets (buf, 60, src) > 0) {
	    if (!strncmp (buf, "STAT: ", 6)) {
		/* Wenn der Status mit gerettet werden soll, dann tue dies */
		if (savestate) {
		    xPuts (buf, dst);
		    while (!cuteol (buf)) {
			if (xGets (buf, 60, src) <= 0) { break; }
			xPuts (buf, dst);
		    }
		} else {
		    while (!cuteol (buf)) {
			if (xGets (buf, 60, src) <= 0) { break; }
		    }
		}
	    } else {
		/* Wenn kein Status vorhanden war, dann schreibe die
		** gelesenen Daten sofort in den Zielfolder
		*/
		xPuts (buf, dst);
	    }
	}
	/* Kopiere den Rest der Daten in den Zielfolder */
	SingleStore (src, dst);
    }
    return ix;
}

/* SaveFailed (from, &to, smtperr, smtpcommand)
**     ffnet - falls notwendig - einen temporren VxMail-Folder und schreibt
**     die gerade ausgewhlte E-Mail aus dem offenen VxMail-Folder `from'
**     hinein; falls der temporre VxMail-Folder neu geffnet wurde, befindet
**     sich das zugehrige `xFile'-handle hinter in dem Ein-/Ausgabeparameter
**     `to'; sonst ist dies das `xFile'-handle, an das die E-Mails gesendet
**     werden.
*/
static void SaveFailed(xFile fld, xFile *sav, char *errmsg, char *failedcmd)
{
    xFile sfh;
    char *tmpfile, temp[1024];
    sprintf (temp, "vxpop-bsmtp.%d", getpid ());
    tmpfile = gconc (__FILE__, __LINE__, prefs.tmpdir, "/", temp, NULL);
    if (!(sfh = *sav)) {
	if (!(sfh = flOpen (tmpfile, "ux", 0600))) {
	    vxlog (0, "opening save file `%s' failed: %s\n",
		      tmpfile, strerror (errno));
	    return;
	}
	flCopyNextUID (fld, sfh);
	*sav = sfh;
	ssav = *sav;
	/* Hiermit spare ich mir das (leidige) `UnlinkSaved' ein. */
	xUnlink (tmpfile);
	gfree (tmpfile);
    }
    xReSet (fld);
    cuteol (failedcmd);
    strcpy (temp, "STAT: ");
    strncpy (temp, errmsg, 1023);
    strncat (temp, " (", 1023 - strlen(temp));
    strncat (temp, failedcmd, 1023 - strlen(temp));
    strncat (temp, ")", 1023 - strlen(temp));
    temp[1023] = '\0';
    /*snprintf (temp, 1024, "STAT: %s (%s)", errmsg, failedcmd);*/
    xPuts (temp, sfh); xPuts ("\r\n", sfh);
    SingleStore (fld, sfh);
}

/* EmergencySave (xFile_hndl)
**     speichert alle E-Mails, die (aufgrund eines Zusammenbruchs der
**     Verbindung) nicht mehr bertragen werden konnten, in einem temporren
**     Folder. Das Filehandle dieses temporren Folders ist der Funktionswert
**     von EmergencySave.
*/
static xFile EmergencySave (xFile fld)
{
    xFile sav;
    int ix;
    char *tmpfile, tmpfname[60];
    sprintf (tmpfname, "vxpop-emergency.%d", getpid());
    tmpfile = gconc (__FILE__, __LINE__, prefs.tmpdir, "/", tmpfname, NULL);
    if (!(sav = flOpen (tmpfile, "ux", 0600))) {
	vxlog (0, "opening emergency save file `%s' failed: %s\n",
		  tmpfile, strerror (errno));
	return (xFile) 0;
    }
    xReSet (fld);
    flCopyNextUID (fld, sav);
    ix = MultiStore (fld, sav, current_msg, 0) - current_msg;
    vxlog (1, "Saved %d unsent messages in emergency folder!\n", ix);
    /* Ein Unix-System ermglicht das Lschen einer Datei, bevor sie
    ** geschlossen wurde. Da ich diese Datei bis zur Beendigung der
    ** Arbeit sowieso nicht mehr schliee, kann ich sie auch gleich
    ** hier lschen.
    */
    xUnlink (tmpfile);
    return sav;
}

static void Fallback (xFile msgs, const char *fbaddr, xFile orig,
		      const char *mailhost, const char *host);

static SIGWRP(time_check)

/* time_check (sig)
**     ist die Signalbehandlungsroutine fr einen abgelaufenen `timeout'.
**     Zuerst werden alle (bis dato) noch nicht gesendeten E-Mails mittels
**     `EmergencySave' in einen temporren Folder gerettet.
**     Anschlieend wird der Folder des entsprechenden Benutzers geleert und
**     mit diesen E-Mails wieder gefllt.
**     Im dritten Schritt werden die unzustellbaren E-Mails ebenfalls wieder in
**     diesen Folder gestellt.
**     Schlielich werden die temporren Folder sowie der Folder des Benutzers
**     wieder freigegeben.
*/
static void time_check (int sig)
{
    /* SIGALRM handler for checking a send timeout */
    int ix;
    xFile sav;
    /* Was mu ich machen?
    ** 1. Retten aller nicht bertragenen E-Mails
    ** 2. Neuerzeugung des Folders mit allen nicht bertragenen E-Mails.
    ** 3. Zurckschreiben der E-Mails mit ungltiger Adresse in den Folder.
    */

    /* Rette alle noch nicht gesendeten E-Mails */
    if (!(sav = EmergencySave (sbFx))) {
	/* Schreibe in die LOG-Datei, da der ursprngliche Folder nicht
	** verndert wird.
	*/
	vxlog (0, "Emergency save failed!\n");
	vxlog (0, "==> The original mailbox will remain unchanged!\n");
	/* Wenn ein temporrer Folder mit fehlgeschlagenen E-Mails existiert,
	** dann wird dieser gelscht.
	*/
	if (ssav) {
	    xClose (ssav);
	    vxlog (1, "temporary file(s) deleted.\n");
	}
	vxlog (0, "+++ `bsmtp' ABORTED! +++\n");
	xClose (sbFx); exit (EX_IOERR);
    }

    /* ffne (und leere) den ursprnglichen Folder */
    xReWrite (sbFx);

    /* Schreibe die nicht gesendeten E-Mails in den ursprnglichen Folder
    ** zurck.
    */
    xReSet (sav);
    ix = MultiStore (sav, sbFx, 0, 0);
    xClose (sav);
    vxlog (1, "Copied %d unsent messages back to folder.\n", ix);

    /* Wenn fehlgeschlagene E-Mails (-> unbekannter Empfnger) da sind,
    ** dann speichere diese ebenfalls wieder im Folder des Benutzers.
    */
    if (ssav) {
	xReSet (ssav);
	/* Falls kein `fallback'-Mechanismus konfiguriert wurde, dann
	** schreibe sie in den ursprnglichen Folder zurck.
	*/
	MultiStore (ssav, sbFx, 0, 0);
	xClose (ssav);
    }

    /* Schliee den (neu erzeugten) Folder */
    xClose (sbFx);
    exit (EX_IOERR);
}

/* Hole eine Zeile des Umschlages; die Zeile wird ohne Zeilenende
** zurckgegeben. Im Fall eines Speicherberlaufes wird das Programm
** abgebrochen.
*/
static char *getenvline (xFile bFx)
{
    /* Der Umschlag der E-Mail mu zunchst gesendet werden; da jedoch u.U.
    ** nderungen notwendig sind (Entfernen des Domainnamens aus der Empfnger-
    ** adresse) mu dies genau berwacht werden. Ich sende also jede Zeile des
    ** Umschlages einzeln.
    */
    size_t lsz = 1024, dist;
    char *res = galloc (__FILE__, __LINE__, 1024), *line;
    line = res;
    while (xGets (line, 1024, bFx) > 0) {
	if (cuteol (line)) { break; }
	lsz += 1024;
	dist = line - res;
	res = gresize (__FILE__, __LINE__, res, lsz);
	line = res + dist;
    }
    return res;
}

/* bsmtp (target_host, bsmtp_folder, thishost)
**     ffnet eine SMTP-Verbindung zum Server `target_host' und bertrgt
**     smtliche in dem VxMail-Folder `bsmtp_folder' stehenden (schon im
**     SMTP-Protokollformat vorliegenden) E-Mails diesen.
**     Der Parameter `thishost' dient der HELLO-Authentifizierung.
**     E-Mails, die nicht erfolgreich zugestellt werden konnten, verbleiben
**     in `bsmtp_folder' (bzw. werden nach beendigung der bertragung dorthin
**     zurckgeschrieben); alle erfolgreich bertragenen E-Mails werden hinge-
**     gen aus `bsmtp_folder' entfernt.
*/
int bsmtp (const char *host, char *_folder, const char *mailhost,
	   const char *fallback, int nodomain)
{
    xFile bFx, sav = (xFile) 0;
    char line[1024], *errmsg, *folder, *prefix, *p, *q;
    int ev, ix, portnum;
    HLSock hlc;
    vxlog (1, "Entering `bsmtp'\n");
#ifdef SETPROCTITLE
    setproctitle ("bsmtp \"%s\" \"%s\" \"%s\"", host, _folder, mailhost);
#endif
    SIGNAL (SIGALRM, time_check);
    alarm (0);
    /* release the terminal?? */
    /*setpgid (getpid (), getpid ());*/
    /*fclose (stdin); fclose (stdout);*/
    /* CHANGE: the default mailbox directory for `bsmtp'-folders is no
    ** longer the list directory; it is now the mailbox base directory!
    */
    folder = abs_path (prefs.mbox_base, _folder);
    vxlog (3, "Folder = `%s'\n", folder);
    if (!(bFx = flOpen (folder, "rX"))) {
	vxerror_at (errno == EBUSY ? EX_TEMPFAIL : EX_OSERR, errno,
		    __FILE__, __LINE__ - 2,
		    "opening mailbox `%s' failed", folder);
    }
    sbFx = bFx;
    vxlog (3, "Folder successfully opened!\n");
    vxlog (3, "Remote HOST = %s\n", host);
    current_msg = 0;
    /* Ich lege fr die Verbindungsaufnahme einen festen `timeout' von 10
    ** Minuten fest.
    */
    alarm (300);
    portnum = LLPort ("smtp", "tcp");
    if (portnum < 0) {
	vxlog (0, "bsmtp: Service (smtp) ==> portnum translation failed!\n");
	xClose (bFx);
	gfree (folder);
	return 1;
    }
    if (!(hlc = HLConnect ("tcp", host, portnum))) {
	vxlog(0, "SMTP connection to `%s' failed: %s\n", host, strerror(errno));
	xClose (bFx);
	gfree (folder);
	return 1;
    }
    vxlog (3, "HLSock initializes for high-level access.\n");
    if (HLhfInit (hlc)) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(HLhfInit)");
    }
    ev = smtp_result (hlc, (char **) NULL);
    if (ev / 100 != 2) {
	vxlog (0, "SMTP-server `%s' didn't send welcome message!\n", host);
	vxlog (0, "SMTP status: %d\n", ev);
	HLClose (hlc);
	xClose (bFx);
	gfree (folder);
	return 2;
    }
    vxlog (3, "SMTP(tcp)-connection to host `%s' established.\n", host);
    HLPuts ("HELO ", hlc); HLPuts (mailhost, hlc); HLPuts ("\r\n", hlc);
    vxlog (5, ">> HELO %s\\r\\n\n", mailhost);
    ev = smtp_result (hlc, (char **) NULL);
    if (ev / 100 != 2) {
	vxlog (0, "SMTP HELO command failed!\n");
	HLClose (hlc);
	xClose (bFx);
	gfree (folder);
	return 3;
    }
    for (ix = 0; xSeek (bFx, (off_t) ix) >= 0; ++ix) {
	current_msg = ix;
	vxlog (5, ">> ");
#if 0
/* bsmtp_timeout should be the response timeout, not the timeout for
** the complete transmission of the e-mail. So i inserted the `alarm'
** system calls into `smtp_result'.
*/
	alarm (prefs.bsmtp_timeout);
#endif
	/* NEU (11.5.2000):
	** Um das Abschneiden des Domainnamens aus der Empfngeradresse nach
	** hier zu verlegen waren massive nderungen notwendig; so habe ich
	** eine neue Funktion geschrieben, die eine Zeile des Umschlages
	** vollstndig in einen dynamisch alloziierten Puffer einliest;
	** dieser Puffer wird (nur in dem Fall, da `nodomain' gesetzt ist)
	** daraufhin untersucht, ob er den RCPT-Befehl enthlt. Falls ja,
	** dann wird nach einem Domainanteil in der Empfngeradresse gesucht;
	** hatte die Empfngeradresse einen Domainanteil, dann wird der Puffer-
	** inhalt ohne diesen ausgegeben.
	** Hatte die Empfngeradresse keinen Domainanteil oder ist `nodomain'
	** nicht gesetzt, dann wird der vollstndige Pufferinhalt ausgegeben;
	** dies geschieht ebenso, wenn die Zeile nicht den `RCPT'-Befehl
	** enthlt. Da die Funktion zum Lesen der Umschlagszeile diese Zeile
	** ohne Zeilenende zurckgibt, mu dieses anschlieend noch nachgesetzt
	** werden.
	** Anschlieend geht es wie gehabt (wie frher) weitergemacht; d.h.,
	** das vom SMTP-Server zurckkommende Ergebnis wird berprft; handelt
	** es sich um eine 2xx, dann wird in dieser Schleife weitergemacht;
	** handelt es sich um eine 3xx (eigentlich nur 354 mglich), dann
	** wird diese Schleife beendet; alles andere entspricht einem Fehler;
	** in diesem Fall wird die fehlgeschlagene E-Mail gerettet und ein
	** `RSET'-Befehl stattdessen ausgegeben, do da wieder ein sauberer
	** bertragungszustand fr die nchsten folgenden E-Mails vorhanden
	** ist.
	*/
	for (;;) {
	    prefix = getenvline (bFx);
	    if (nodomain && !lccmp (prefix, "RCPT TO:", 8)) {
		p = strchr (&prefix[8], '@');
		unlessnull (p) {
		    q = strpbrk (&p[1], (prefix[8] == '<' ? ">" : " \t"));
		    ifnull (q) {
			*p = '\0'; HLPuts (prefix, hlc);
			if (prefix[8] == '<') { HLPuts (">", hlc); }
		    } else if (q - p < 0) {
			HLPuts (prefix, hlc);
		    } else {
			*p = '\0'; HLPuts (prefix, hlc); HLPuts (q, hlc);
		    }
		    vxlog (5, "%s%s\n", prefix, (q == NULL ? ">" : q));
		} else {
		    HLPuts (prefix, hlc);
		    vxlog (5, "%s\n", prefix);
		}
	    } else if (!lccmp (prefix, "STAT: ", 6)) {
		/* skip the `STAT: ...' line, VxPOP inserted previously... */
		vxlog (2, "Re-sending previously failed message...\n");
		gfree (prefix); prefix = NULL; continue;
	    } else {
		HLPuts (prefix, hlc);
		vxlog (5, "%s\n", prefix);
	    }
	    HLPuts ("\r\n", hlc);
	    ev = smtp_result (hlc, &errmsg);
	    if (ev / 100 == 3) { gfree (errmsg); gfree (prefix); break; }
	    if (ev / 100 != 2) {
		SaveFailed (bFx, &sav, errmsg, prefix);
		vxlog (0, "SMTP transmission of message %d failed!\n", ix);
		vxlog (0, "SMTP-Status is: %d.\n", ev);
		vxlog (0, "Re-initializing the connection (RSET).\n");
		HLPuts ("RSET\r\n", hlc);
		(void) smtp_result (hlc, (char **) NULL);
		gfree (errmsg); gfree (prefix);
		goto MSG_DONE;
	    }
	    gfree (errmsg);
	    gfree (prefix); prefix = NULL;
	}
	while (xGets (line, 1024, bFx) > 0) {
	    HLPuts (line, hlc);
	    vxlog (5, "%s", line);
	    if (cuteol (line)) {
		vxlog (5, ">> ");
		if (!strcmp (line, ".")) { break; }
	    }
	}
	if (strcmp (line, ".")) {
	    if (!cuteol (line)) { HLPuts ("\r\n", hlc); vxlog (5, "\n"); }
	    HLPuts (".\r\n", hlc);
	    vxlog (5, ">> .\\r\\n\n");
	}
	ev = smtp_result (hlc, &errmsg);
	if (ev / 100 != 2) {
	    SaveFailed (bFx, &sav, errmsg, "DATA");
	    vxlog (0, "SMTP transmission of message %d failed!\n", ix);
	}
	gfree (errmsg);
    MSG_DONE:
	;
    }
    HLPuts ("QUIT\r\n", hlc);
    vxlog (5, ">> QUIT\\r\\n\n");
    ev = smtp_result (hlc, (char **) NULL);
    if (ev / 100 != 2) {
	vxlog (0, "SMTP QUIT command failed "
		  "- nevertheless closing the connection!\n");
    }
    HLClose (hlc);
    alarm (0);
    vxlog (3, "SMTP-transfer completed -- cleaning up folder.\n");

    /* Entleere den Folder */
    xReWrite (bFx);
    if (sav) {
	xReSet (sav);
	Fallback (sav, fallback, bFx, mailhost, host);
	xClose (sav);
    }
    xClose (bFx);
    gfree (folder);
    vxlog (1, "Leaving `bsmtp'!\n");
    return 0;
}

static __inline__ void send_line (xFile msgs, HLSock hlc, char line[1024])
{
    HLPuts (line, hlc);
    vxlog (5, ">> %s", line);
    while (!cuteol (line)) {
	if (xGets (line, 1024, msgs) <= 0) { break; }
	HLPuts (line, hlc);
	vxlog (5, line);
    }
}

static xFile smsgs = (xFile) 0, sorig = (xFile) 0;

static SIGWRP(time_fallback)

static void time_fallback (int sig)
{
    int ix = MultiStore (smsgs, sorig, current_msg, 0);
    vxlog (0, "stored %d messages back into original folder\n", ix-current_msg);
    xClose (smsgs); xClose (sorig);
    exit (EX_IOERR);
}

static void SendBack (xFile msgs, xFile orig, const char *mailhost);

static int skipeol (xFile msgs, char *buf)
{
    char line[1024];
    if (cuteol (buf)) { return 0; }
    while (xGets (line, 1024, msgs) > 0 && !cuteol (line));
    return 1;
}

static void Fallback (xFile msgs, const char *fbaddr, xFile orig,
		      const char *mhost, const char *host)
{
    HLSock hlc;
    int ev, ix, portnum;
    char line[1024], rorig[1024];
    if (fbaddr == NULL) {
	/* Hier mu der Default-Mechanismus zur Behandlung von E-Mails mit
	** unbekanntem Empfnger installiert werden.
	** Es wird ein umgekehrter Sendevorgang eingeleitet, d.h. die
	** Bedeutung der `MAIL FROM:'- und `RCPT TO:'-Zeilen werden vertauscht.
	** Als Relay hierfr wird der in der VxMail-Konfiguration fr die
	** `LLTO'-Eintrge zustndige SMTP-Server verwendet. Wenn dieser nicht
	** existiert, dann werden die E-Mails zurckgespeichert. Alle E-Mails
	** werden vor dem Zurckschicken um Fehlerinformationen erweitert.
	*/
	SendBack (msgs, orig, mhost);
	return;
    }
    SIGNAL (SIGALRM, time_fallback);
    alarm (300);
    portnum = LLPort ("smtp", "tcp");
    if (portnum < 0) {
	vxlog (0, "fallback: Service (smtp) ==> portnum translation failed!\n");
	ix = MultiStore (msgs, orig, 0, 0);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    if (!(hlc = HLConnect ("tcp", host, portnum))) {
	vxlog(0, "SMTP fallback connection to `%s' failed: %s\n",
		 host, strerror(errno));
	ix = MultiStore (msgs, orig, 0, 0);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    vxlog (3, "HLSock initializes for high-level access.\n");
    if (HLhfInit (hlc)) {
	int _saverr = errno;
	vxlog (0, "FATAL! Unable to initialize high level function support.\n");
	ix = MultiStore (msgs, orig, 0, 0); alarm (0);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	errno = _saverr;
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(HLhfInit)");
    }
    ev = smtp_result (hlc, (char **) NULL);
    if (ev / 100 != 2) {
	ix = MultiStore (msgs, orig, 0, 0); HLClose (hlc);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	vxlog (0, "SMTP-server `%s' didn't send welcome message!\n", host);
	return;
    }
    vxlog (3, "SMTP(tcp)-fallback connection to host `%s' established.\n",
	      host);

    /* Sende den HELLO (HELO)-Befehl.  */
    HLPuts ("HELO ", hlc); HLPuts (mailhost (), hlc); HLPuts ("\r\n", hlc);
    if ((ev = smtp_result (hlc, (char **) NULL)) / 100 != 2) {
	ix = MultiStore (msgs, orig, 0, 0); HLClose (hlc);
	vxlog (0, "Hello command failed: SMTP status: %d\n", ev);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    /* Rette die Foldervariablen `msgs' und `orig' in globale Variable
    ** fr eine `timeout'-Behandlung
    */
    smsgs = msgs;
    sorig = orig;
    for (ix = 0; xSeek (msgs, (off_t) ix) >= 0; ++ix) {
	/* Rette die Meldungsnummer in eine globale Variable fr eine
	** `timeout'-Behandlung.
	*/
	current_msg = ix;
#if 0
/* 26.5.2000 bj: see above */
	alarm (prefs.bsmtp_timeout);
#endif

	/* berspringe eine (evtl. vorhandene) `STAT: '-Zeile */
	xGets (line, 1024, msgs);
	if (!strncmp (line, "STAT: ", 6)) {
	    skipeol (msgs, line);
	    xGets (line, 1024, msgs);
	}

	/* bertrage die `MAIL FROM:'-Zeile */
	send_line (msgs, hlc, line);
	ev = smtp_result (hlc, (char **) NULL);
	if (ev / 100 != 2) {
	    ix = MultiStore (msgs, orig, ix, 0); HLClose (hlc);
	    vxlog (0, "Sending mail failed: SMTP status: %d\n", ev);
	    vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	    return;
	}
	/* berspringe die `RCPT TO:'-Zeile, merke dir aber den ursprnglichen
	** Empfnger
	*/
	xGets (rorig, 1024, msgs);
	skipeol (msgs, rorig);

	/* bertrage die alternative Adresse in einer neu generierten
	** `RCPT TO:'-Zeile
	*/
	HLPuts ("RCPT TO:<", hlc); HLPuts (fbaddr, hlc); HLPuts (">\r\n", hlc);
	vxlog (5, ">> RCPT TO:<%s>\n", fbaddr);
	if ((ev = smtp_result (hlc, (char **) NULL)) / 100 != 2) {
	    ix = MultiStore (msgs, orig, ix, 0);
	    HLClose (hlc);
	    vxlog (0, "RCPT command failed: SMTP status: %d\n", ev);
	    vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	    return;
	}

	/* bertrage die `DATA'-Zeile */
	xGets (line, 1024, msgs);
	send_line (msgs, hlc, line);
	if ((ev = smtp_result (hlc, (char **) NULL)) / 100 != 3) {
	    ix = MultiStore (msgs, orig, ix, 0);
	    HLClose (hlc);
	    vxlog (0, "DATA command failed: SMTP status: %d\n", ev);
	    vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	    return;
	}

	/* Nun mu die `From ...'-Zeile bergeben werden. */
	xGets (line, 1024, msgs);
	if (!strncmp (line, "From ", 5)) {
	    send_line (msgs, hlc, line);
	    xGets (line, 1024, msgs);
	    while (*line == ' ' || *line == '\t') {
		vxlog (5, ">> ");
		send_line (msgs, hlc, line);
		xGets (line, 1024, msgs);
	    }
	}
	HLPuts ("X-Orig-RCPT: ", hlc); HLPuts (&rorig[8], hlc);
	HLPuts ("\r\n", hlc);
	vxlog (5, "X-Orig-RCPT: %s\n", &rorig[8]);
	HLPuts (line, hlc);
	vxlog (5, "%s", line);
	while (strcmp (line, ".\r\n")) {
	    xGets (line, 1024, msgs);
	    HLPuts (line, hlc);
	    vxlog (5, "%s", line);
	}
	if ((ev = smtp_result (hlc, (char **) NULL)) / 100 != 2) {
	    ix = MultiStore (msgs, orig, ix, 0);
	    HLClose (hlc);
	    vxlog (0, "Sending e-mail no %d failed: SMTP status: %d\n",
		    current_msg, ev);
	    vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	    return;
	}
    }
    HLPuts ("QUIT\r\n", hlc);
    vxlog (5, ">> QUIT\n");
    
    if ((ev = smtp_result (hlc, (char **) NULL)) / 100 != 2) {
	vxlog (0, "WARNING! QUIT command failed (all messages sent)!\n");
    }
    HLClose (hlc);
}

/*static const char *smtp_errmsg (int code);*/

static SIGWRP(time_sendback)

static void time_sendback (int sig)
{
    int ix = MultiStore (smsgs, sorig, current_msg, 0);
    vxlog (0, "stored %d messages back into original folder\n", ix-current_msg);
    xClose (smsgs); xClose (sorig);
    exit (EX_IOERR);
}

static char unskipped[1024];
static __inline__ int skip_from (xFile msgs)
{
    if (xGets (unskipped, 1024, msgs) > 0) {
	if (strncmp (unskipped, "From ", 5)) { return 1; }
	while (!cuteol (unskipped) && xGets (unskipped, 1024, msgs) >= 0);
	while (xGets (unskipped, 1024, msgs) > 0) {
	    if (*unskipped != ' ' && *unskipped != '\t') { return 1; }
	    while (!cuteol (unskipped) && xGets (unskipped, 1024, msgs) >= 0);
	}
    }
    return 0;
}

static __inline__ void SkipStat (xFile msgs, xFile orig)
{
    char line[1024];
    if (xGets (line, 1024, msgs) > 0) {
	if (strncmp (line, "STAT: ", 6)) { xPuts (line, orig); }
    }
}

static void SplitStat (char *stat, char *smtperr, char *smtpcmd)
{
    char *p, *q;
    int pc, len;
    cuteol (stat);
    p = strrchr (stat, ')');
    if (p != NULL) {
	pc = 1; q = p;
	while (p != stat && pc > 0) {
	    --p;
	    if (*p == ')') ++pc;
	    if (*p == '(') --pc;
	}
	if (p != stat && --p != stat) {
	    len = p - stat; if (len > 1023) { len = 1023; }
	    memcpy (smtperr, stat, len);
	    smtperr[len] = '\0';
	    p += 2;
	    len = q - p; if (len > 1023) { len = 1023; }
	    memcpy (smtpcmd, p, len);
	    smtpcmd[len] = '\0';
	    return;
	}
    }
    len = strlen (stat); if (len > 1023) { len = 1023; }
    memcpy (smtperr, stat, len);
    smtperr[len] = '\0';
    strcpy (smtpcmd, ">>unknown<<");
}

static void SendBack (xFile msgs, xFile orig, const char *mailhost)
{
    HLSock hlc;
    int ix, lasteol, preveol;
    char recipient[1024], sender[1024], smtperr[1024], failed[1024], line[1024];
    char *adms;
    int ev, portnum;
    if (prefs.nonlocalto == NULL) {
	ix = MultiStore (msgs, orig, 0, 0);
	vxlog (1, "no relay host configured; messages stored back in the\n");
	vxlog (1, "original folder.\n");
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    smsgs = msgs; sorig = orig; current_msg = 0;
    SIGNAL (SIGALRM, time_sendback);
    alarm (300);
    portnum = LLPort ("smtp", "tcp");
    if (portnum < 0) {
	vxlog (0, "sendback: Service (smtp) ==> portnum translation failed!\n");
	ix = MultiStore (msgs, orig, 0, 0);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    if (!(hlc = HLConnect ("tcp", prefs.nonlocalto, portnum))) {
	vxlog(0, "SMTP fallback connection to `%s' failed: %s\n",
		 prefs.nonlocalto, strerror(errno));
	ix = MultiStore (msgs, orig, 0, 0);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    vxlog (3, "SendBack: HLSock initializes for high-level access.\n");
    if (HLhfInit (hlc)) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(HLhfInit)");
    }
    ev = smtp_result (hlc, (char **) NULL);
    if (ev / 100 != 2) {
	ix = MultiStore (msgs, orig, 0, 0); HLClose (hlc);
	vxlog (0, "SMTP-server `%s' didn't send welcome message!\n",
		  prefs.nonlocalto);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    vxlog (3, "SMTP(tcp)-fallback connection to host `%s' established.\n",
	      prefs.nonlocalto);
    HLPuts ("HELO ", hlc); HLPuts (mailhost, hlc); HLPuts ("\r\n", hlc);
    vxlog (5, ">> HELO %s\n", mailhost);
    ev = smtp_result (hlc, (char **) NULL);
    if (ev / 100 != 2) {
	ix = MultiStore (msgs, orig, 0, 0); HLClose (hlc);
	vxlog (0, "HELO command failed; SMTP-status: %d\n", ev);
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
	return;
    }
    for (ix = 0; xSeek (msgs, (off_t) ix) >= 0; ++ix) {
	/* Hier mssen jetzt die einzelnen E-Mails zurckgesendet werden.
	** Dabei gilt: eine fehlgeschlagene E-Mail wird nicht mehr gespeichert;
	** sie wird verworfen!
	** Ich gehe davon aus, da der SMTP-Status in einer Zeile der Form:
	** `STAT: <code>' im Folder gespeichert wurde, wobei <code> der drei-
	** stellige SMTP-Statuscode ist.
	*/
	current_msg = ix;
#if 0
/* 26.5.2000 bj: see above */
	alarm (prefs.bsmtp_timeout);
#endif
	xGets (line, 1024, msgs);
	if (!strncmp (line, "STAT: ", 6)) {
	    /* Die `STAT: '-Zeile enthlt die SMTP-Rckmeldung und die Ursache
	    ** fr diese. Ich bentige beide getrennt voneindander, mu sie
	    ** daher splitten.
	    */
	    skipeol (msgs, line);
	    SplitStat (&line[6], smtperr, failed);
	    xGets (recipient, 1024, msgs);
	} else {
	    strcpy (smtperr, "550 unknown recipient");
	    strcpy (recipient, line);
	}
	if (skipeol (msgs, recipient)) {
	    vxlog (0, "WARNING! `MAIL FROM:'-line too long!\n");
	}
	xGets (sender, 1024, msgs);
	if (skipeol (msgs, sender)) {
	    vxlog (0, "WARNING! `RCPT TO:'-line too long!\n");
	}

	/* berspringen des DATA-Befehls */
	xGets (line, 1024, msgs);

	/* Ausgabe des Absenders der E-Mail - in diesem Fall das
	** `local mail delivery subsystem' (VxMail).
	*/
	HLPuts ("MAIL FROM:", hlc);
	adms = gvarins (__FILE__, __LINE__,
			prefs.error_sender, "MAILHOST", mailhost);
	HLPuts (adms, hlc);
	HLPuts ("\r\n", hlc);
	vxlog (5, ">> MAIL FROM:%s\n", adms);
	ev = smtp_result (hlc, (char **) NULL);
	if (ev / 100 != 2) {
	    /* Wenn das fehlschlgt, dann schreibe die E-Mail (ohne die von
	    ** diesem Programm zwischenzeitlich eingefgte `STAT: '-Zeile)
	    ** zurck in den Folder, aus dem sie ursprnglich kam.
	    */
	    gfree (adms);
	    SkipStat (msgs, orig); SingleStore (msgs, orig);
	    vxlog (0, "`MAIL FROM:' failed - skipping message...\n");
	    HLPuts ("RSET\r\n", hlc);
	    ev = smtp_result (hlc, (char **) NULL);
	    continue;
	}

	/* Ausgabe des neuen Empfngers - dies ist der ursprngliche
	** Absender.
	*/
	HLPuts ("RCPT TO:", hlc); HLPuts (&recipient[10], hlc);
	HLPuts ("\r\n", hlc);
	vxlog (5, ">> RCPT TO:%s\n", &recipient[10]);
	ev = smtp_result (hlc, (char **) NULL);
	if (ev / 100 != 2) {
	    /* Wenn das fehlschlgt, dann schreibe die E-Mail (ohne die von
	    ** diesem Programm zwischenzeitlich eingefgte `STAT: '-Zeile)
	    ** zurck in den Folder, aus dem sie ursprnglich kam.
	    */
	    gfree (adms);
	    SkipStat (msgs, orig); SingleStore (msgs, orig);
	    vxlog (0, "`RCPT TO:' failed - skipping message...\n");
	    HLPuts ("RSET\r\n", hlc);
	    ev = smtp_result (hlc, (char **) NULL);
	    continue;
	}

	/* Beginnen der eigentlichen Nachricht (DATA-Befehl) */
	HLPuts ("DATA\r\n", hlc);
	vxlog (5, ">> DATA\n");
	ev = smtp_result (hlc, (char **) NULL);
	if (ev / 100 != 3) {
	    gfree (adms);
	    SkipStat (msgs, orig); SingleStore (msgs, orig);
	    vxlog (0, "`DATA' failed - skipping message...\n");
	    HLPuts ("RSET\r\n", hlc);
	    ev = smtp_result (hlc, (char **) NULL);
	    continue;
	}
	/* Zunchst mal meinen eigenen Kram */
	HLPuts ("From <BSMTP@", hlc); HLPuts (mailhost, hlc);
	HLPuts ("> (VxPOP)\r\n", hlc);
	vxlog (5, "From <BSMTP@%s> (VxPOP)\n", mailhost);
	HLPuts ("From: ", hlc); HLPuts (adms, hlc); HLPuts ("\r\n", hlc);
	vxlog (5, "From: %s\n", adms);
	gfree (adms);
	HLPuts ("To: ", hlc); HLPuts (&recipient[10], hlc);
	HLPuts ("\r\n", hlc);
	vxlog (5, "To: %s\n", &recipient[10]);
	HLPuts ("Return-Path: <MAILER-DAEMON@", hlc); HLPuts (mailhost, hlc);
	HLPuts (">\r\n", hlc);
	vxlog (5, "Return-Path: <MAILER-DAEMON@%s>\n", mailhost);
	HLPuts ("X-Orig-RCPT: ", hlc); HLPuts (&sender[8], hlc);
	HLPuts ("\r\n", hlc);
	/* Dann (mit einem Spezialtag) den ursprnglich angedachten Empfnger
	** dieser E-Mail.
	*/
	vxlog (5, "X-Orig-RCPT: %s\n", &sender[8]);
	/* Dann das `Subject:', die fehlerhafte Zieladresse und den
	** SMTP-Fehlercode.
	*/
	HLPuts ("Subject: MAIL FAILED - returning to sender\r\n", hlc);
	vxlog (5, "Subject: MAIL FAILED - returning to sender\n");
	HLPuts ("\r\n|------------------------- Failed address follows: "
		"-----------------------|\r\n", hlc);
	vxlog (5, "|------------------------- Failed address follows: "
		  "-----------------------|\r\n");
	HLPuts ("  ", hlc);
	/* Die anzugebende Zieladresse mu entweder der Inhalt von `failed'
	** (falls `failed' mit `MAIL FROM:' oder `RCPT TO:' beginnt) oder
	** die ursprngliche Empfngeradresse sein.
	*/
	if (!lccmp (failed, "MAIL FROM:", 10)) {
	    HLPuts (&failed[10], hlc);
	    vxlog (5, "  %s\n", &failed[10]);
	} else if (!lccmp (failed, "RCPT TO:", 8)) {
	    HLPuts (&failed[8], hlc);
	    vxlog (5, "  %s\n", &failed[8]);
	} else {
	    HLPuts (&sender[8], hlc);
	    vxlog (5, "  %s\n", &sender[8]);
	}
	HLPuts ("\r\n|------------------------- SMTP-Messages: ---------"
		"-----------------------|\r\n", hlc);
	vxlog (5, "|------------------------- SMTP-Messages: ---------"
		  "-----------------------|\r\n");
	HLPuts (smtperr, hlc); HLPuts ("\r\n", hlc);
	vxlog (5, "%s\n", smtperr);
	HLPuts ("\r\n|------------------------- Original message: ------"
		"-----------------------|\r\n", hlc);
	vxlog (5, "|------------------------- Original message: ------"
		  "-----------------------|\r\n");
	/* Zum Schlu die ursprngliche E-Mail (ohne die `From '-Zeile). */
	if (skip_from (msgs)) { HLPuts (unskipped, hlc); }
	vxlog (5, ">> ");
	lasteol = 1;
	while (xGets (line, 1024, msgs) > 0) {
	    preveol = lasteol;
	    HLPuts (line, hlc);
	    lasteol = cuteol (line);
	    vxlog (5, (lasteol ? "%s\n" : "%s"), line);
	    /* Nur dann, wenn dies eine vollstndige Zeile - und nicht der Rest
	    ** einer Zeile aus einem der vorherigen Durchlufe - ist, dann darf
	    ** nach der Leerzeile, die das Ende des Headers darstellt,
	    ** gefragt werden. Nach `lasteol' mu ich hier jedoch nicht mehr
	    ** fragen, weil eine leere Zeile IMMEr bedeutet, da ein Zeilenende
	    ** gefunden wurde.
	    */
	    if (preveol && !*line) { break; }
	}
	if (prefs.biirm) {
	    /* Nur dann, wenn das Versenden des Rumpfes der ursprnglichen
	    ** E-Mail beim Versenden der Fehler-E-Mail ausdrcklich zugelassen
	    ** wurde, dann wird auch noch der Rest der E-Mail verschickt.
	    */
	    while (xGets (line, 1024, msgs) > 0) {
		HLPuts (line, hlc);
		lasteol = cuteol (line);
		vxlog (5, (lasteol ? "%s\n" : "%s"), line);
	    }
	    if (!strcmp (line, ".")) {
		HLPuts (".\r\n", hlc);
		vxlog (5, ".\n");
	    }
	} else {
	    /* Ansonsten wird einfach nur eine Zeile versandt, die ausdrckt,
	    ** da kein Rumpf enthalten ist.
	    */
	    HLPuts ("[no body included]\r\n.\r\n", hlc);
	    vxlog (5, "[no body included]\n.\n");
	}
	ev = smtp_result (hlc, (char **) NULL);
	if (ev / 100 != 2) {
	    /* Wenn das fehlschlgt, dann schreibe die E-Mail (ohne die von
	    ** diesem Programm zwischenzeitlich eingefgte `STAT: '-Zeile)
	    ** zurck in den Folder, aus dem sie ursprnglich kam.
	    */
	    SkipStat (msgs, orig); SingleStore (msgs, orig);
	    vxlog (0, "sending message failed - skipping message...\n");
	    HLPuts ("RSET\r\n", hlc);
	    ev = smtp_result (hlc, (char **) NULL);
	}
    }

    /* Wenn alle fehlgegangenen E-Mails zurckgeschickt wurden, dann beende
    ** die SMTP-Sitzung.
    */
    HLPuts ("QUIT\r\n", hlc);
    vxlog (5, ">> QUIT\n");
    ev = smtp_result (hlc, (char **) NULL);
    if (ev / 100 != 2) {
	vxlog (0, "WARNING! `QUIT' SMTP-command failed; SMTP-status: %d\n",
		  ev);
    }
}
