/* mbox.h
**
** Interface file for `MBOX' - the abstract datatype for accessing
** (selectable) mailboxes.
**
*/

/* Abstrakter Datentyp fr den Zugriff auf Mailboxen */

typedef struct _mbox_handle *MBOX;

/* Verbinden mit einer Mailbox
** Das Ergebnis ist entweder ein MBOX-Zeiger (wenn alles geklappt hat) oder NULL
** - wenn's ein Fehlschlag war.
** Hinweis! Der Benutzer soll die Mglichkeit haben, sich verschiedene Mailboxen
** selbst anzulegen. Es soll im wesentlichen nur eine Standard-Mailbox geben.
** Diese Mailbox hat keinen Namen; sie wird durch einen Wert `NULL' fr das
** Argument `mbox' definiert.
*/

MBOX mboxOpen (char *domain, char *user, char *pfx, char *mbox, int mbox_flags);

/* Wechseln der aktiven Mailbox beim gleichen Benutzer.
** Hinweis: Wenn die betreffende Mailbox noch nicht existiert, dann wird sie
** automatisch angelegt.
*/

int mboxSelect (MBOX mbx, char *mbox);

/* Schreiben von E-Mail-Daten in die aktive Mailbox.
** Das Ergebnis ist die Anzahl an Bytes, die neu hinzugefgt wurden - oder -1,
** wenn etwas schiefgegangen ist.
** Hinweis: Wenn die betreffende Mailbox noch nicht existiert, dann wird sie
** automatisch angelegt.
*/

ssize_t mboxWrite (MBOX mbx, void *buffer, size_t buffersize);

/* Beenden des stckweisen Schreibens in die aktive Mailbox.
** Das Ergebnis ist die Nummer der neuen E-Mail.
*/

int mboxStore (MBOX mbx);

/* Abbruch des stckweisen Schreibens in die aktive Mailbox. */

int mboxCancel (MBOX mbx);

/* Lschen einer E-Mail aus der aktiven Mailbox */

int mboxDelMsg (MBOX mbx, int msgno);

/* Stckweises laden einer E-Mail aus der aktiven Mailbox.
** Das Ergebnis ist -1 bei einem Fehler, 0 beim Erreichen des Endes der
** E-Mail und die Anzahl gelesener Bytes sonst.
*/

ssize_t mboxRetrieve (MBOX mbx, int msgno, void *buffer, size_t buffersize);

/* Abbruch des stckweisen Lesens aus einer Mailbox */

int mboxReset (MBOX mbx);

/* Lesen der Meta-Daten (Nummer, Gre) einer E-Mail innerhalb einer Mailbox
** Ergebnix ist 0, falls OK, 1 falls keine weitere E-Mail mehr vorhanden und
** -1 im Fehlerfall.
*/

int mboxMsgData (MBOX mbx, int seqno, int *msgno, size_t *msgsize);

/* Lschen einer bestimmten Mailbox; anschlieend wird die Standard-Mailbox zur
** aktiven Mailbox - auch wenn es diese ist, die gerade mit dieser Routine
** gelscht wurde.
** Hinweis! Fehler werden nur dann zurckgegeben, wenn das lschen der Mailbox
** wirklich nicht mglich war - nicht jedoch, wenn sie nur nicht existiert hat.
** (nicht-existierende Mailboxen weren AUTOMATISCH ANGELEGT!)
*/
int mboxDelete (MBOX mbx, char *mailbox);

/* Alle bisherigen Lschoperationen auf der aktiven Mailbox rckgngig machen. */
int mboxUndelete (MBOX mbx);

/* Verschieben einer E-Mail aus der aktiven Mailbox in eine andere - angegebene.
** Wenn die betreffende E-Mail nicht existiert, dann wird ein Fehler zurckgegeben.
** Wenn aktive und Ziel-Mailbox bereinstimmen wird (auer in dem Fall, da die
** ausgewhlte E-Mail nicht existiert !!!) nichts getan.
** Bei erfolgreicher Durchfhrung ist das Ergebnis die Nummer der E-Mail in der
** neuen Mailbox.
** Hinweis! Wenn eine der beiden Mailboxen nicht existiert, dann wird sie angelegt.
*/
int mboxMove (MBOX mbx, int msgno, char *mailbox)

/* berprfen, wieviel Platz noch zur Verfgung steht (Mail-Quota!).
** Hinweis! Zur berprfung werden smtliche Mailboxen des jeweiligen Benutzers
** herangezogen. Die betreffende Operation ist als m.u. etwas Zeitaufwendig und
** sollte nicht allzu hufig ausgefhrt werden.
*/

off_t mboxQuota (MBOX mbx, off_t maxsize);

/* Schlieen der Verbindung zu allen Mailboxen des jeweiligen Benutzers.
** Noch anstehende Transaktionen (entfernen als gelscht markierter E-Mails)
** werden vorher noch durchgefhrt.
*/

int mboxClose (MBOX mbx);

